// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// nbdy.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// See the detailled comments for how to use this code. Some abbreviations:    |
//                                                                             |
// R:       return value                                                       |
// I:       function argument is input                                         |
// O:       function argument is output                                        |
// I/O:     function argument is input & output                                |
// [I]:     function argument is optional                                      |
//                                                                             |
// Note that in case of optional function arguments, one may only omit the     |
// last one (and the second but last if the last is already omitted ...).      |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// Note that it is often more convenient to use the wrapper provided by the    |
// classes defined in files yanc.h and ssgs.h                                  |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
//  ON THE TIME-STEPPING SCHEMES                                               |
//  ============================                                               |
//                                                                             |
//  As of February 2002, we have abondoned the old schemes and introduced a    |
//  new scheme instead. It is based on the following prescription (Tremaine):  |
//                                                                             |
//    x  ->  x + h/2 * v    and compute the acceleration a(x)                  |
//    v  ->  v + h/2 * a(x)                                                    |
//    h  ->  h'             based on s(h,h') = tau(x,v,Phi,a,...)              |
//    v  ->  v + h/2 * a(x)                                                    |
//    x  ->  x + h/2 * v                                                       |
//                                                                             |
//  where s(h,h') is a symmetric function that satisfies s(h,h)=h, while tau() |
//  gives the ideal time step. This scheme can also be written as              |
//                                                                             |
//    v  ->  v + h/2 * a(x)                                                    |
//    x  ->  x + h   * v    and compute the acceleration a(x)                  |
//    h  ->  h'                                                                |
//                                                                             |
//  Currently, we are allowing for the following options                       |
//                                                                             |
//    tau = f_a / |a|                (as in GADGET, Springel et al)            |
//    tau = f_p / |Phi|                                                        |
//    tau = f_c * sqrt(|Phi|) / a    (as in PKDGRAV, Stadel)                   |
//                                                                             |
//  and combinations of these (whereby the minimum tau is used).               |
//                                                                             |
//  Note that with this method the time step h may not be a monotonic function |
//  of time even when tau is. In particular, if tau=const but h<tau, we have   |
//  an oscillation between a value h<tau and a value h'>tau.                   |
//                                                                             |
//  Using the Block Step Scheme                                                |
//  ---------------------------                                                |
//                                                                             |
//  In our application of the above to the blockstep scheme we specify s(h,h') |
//  to be the geometric mean and replace tau by the integer power of sqrt(2)   |
//  nearest (in the log) to tau. This results in the following scheme.         |
//                                                                             |
//  h' = h/2  if  h > tau*2^( 1/4)                                             |
//  h' = 2 h  if  h < tau*2^(-1/4) and  h'= 2 h is possible with the blockstep |
//  h' = h    otherwise                                                        |
//                                                                             |
//  Note that our restriction to change the time step by no more than a factor |
//  of two results in violation of the time reversibility. However, for        |
//  carefully chosen time steps, the necessity to change by a factor of 4 or   |
//  more should hardly occur.                                                  |
//                                                                             |
//  For the initial h we take the integer power of 2 that is nearest to tau.   |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_nbdy_h
#define included_nbdy_h

#ifndef included_falcON__h
#  include <falcON.h>
#endif

#ifndef included_pext_h
#  include <nbdy/pext.h>
#endif

#ifndef included_step_h
#  include <nbdy/step.h>
#endif

#ifndef included_ctime
#  include <ctime>
#  define included_ctime
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::basic_nbody                                                    
  //////////////////////////////////////////////////////////////////////////////
  class basic_nbody : 
    protected falcON
  {
    basic_nbody(const basic_nbody&);               // not implemented           
    basic_nbody& operator= (const basic_nbody&);   // not implemented           
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
  public:
    enum soft_type {                               // type of softening         
      global_fixed        = 0,                     // globally time-constant    
      individual_fixed    = 1,                     // individual but time-const 
      individual_adaptive = 2                      // individual and adaptive   
    };
    //--------------------------------------------------------------------------
  protected:
    enum how_to_make_tree {
      grow  = 1,
      reuse = 2,
      free  = 4
    };
    //--------------------------------------------------------------------------
    typedef real tensor[NDIM][NDIM];               // not necessarily symmetric 
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
  protected:
    const bodies   *BODIES;
    const potential*PEX;
    soft_type       SOFTENING;                     // type of softening method  
    real            TINI;                          // initial simulation time   
  private:
    mutable clock_t C_OLD;
    uint            NSOFT;
    uint            NREF;
    const uint      NREUSE;
    uint            N_REUSED;
    const uint      NCRIT;                         // max # bodies in cell      
    real            EMAX;                          // max softening length      
          uint      DIRECT[4];                     // direct summation          
#ifdef __GRAPE
    const bool      GRAPE;                         // using GRAPE?              
#endif
    mutable real    M,Ktot,Utot,TU;                // mass, kin & pot E, vir rat
    mutable amom    L;                             // total angular momentum    
    mutable tensor  KT,WT;                         // kin & pot energy, AM tens 
    mutable vect    CMX,CMV;                       // center of mass pos & vel  
  protected:
    mutable bool    DIAG;                          // diagnose up-to-date ?     
    mutable real    CPU_BUILD,                     // time for falcON::grow etc 
                    CPU_GRAV,                      // time for gravity etc      
                    CPU_PEX,                       // time for ext potential    
                    CPU_DENS,                      // time for density estimates
                    CPU_EPSI,                      // time for setting eps_i    
                    CPU_STEP,                      // total time for a longstep 
                    CPU_TOTAL;                     // total time so far         
    //--------------------------------------------------------------------------
    // protected methods                                                        
    //--------------------------------------------------------------------------
  protected:
    void diagnose             () const;
    void update_diagnostics   () const;
    real trace                (const tensor&) const;
    void set_tree             (const vect* =0, const how_to_make_tree=free);
    void set_forces           ();
    void reset_cpus           () const;
    void update_cpu_total     () const;
    void setup_eps            (body&) const;
    void set_softening_lengths();
    void estimate_mass_density(const bool);
    //--------------------------------------------------------------------------
    // abstract methods                                                         
    //--------------------------------------------------------------------------
  public:
    virtual const real& time() const=0;            // actual time               
    virtual void full_step  (const vect* =0)=0;    // one full time step        
    virtual void long_stats (std::ostream&) const=0;
    virtual void short_stats(std::ostream&) const=0;
    virtual void short_stat_head (std::ostream&) const=0;
    virtual void short_stat_line (std::ostream&) const=0; 
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
    void estimate_mass_densities(                  // estimate every bodies rho 
				 const bool=true); //[I: add contrib of Pot_ext]
    void estimate_surf_densities();                // estimate every bodies SD  
    const real& kin_energy      () const;          // T   = kinetic energy      
    const real& pot_energy      () const;          // U   = potential energy    
    const real  total_energy    () const;          // T+U = total energy        
    const real& virial_ratio    () const;          // -T/U                      
    const amom& total_angmom    () const;          // total angular momentum    
    const vect& total_momentum  () const;          // total momentum            
    const vect& center_of_mass  () const;          // center of mass            
    const real& kin_energy      (const indx, const indx) const; // K_ij         
    const real& pot_energy      (const indx, const indx) const; // W_ij         
    const real  as_angmom       (const indx, const indx) const; // anti-sym AM  
    const real& cpu_build       () const { return CPU_BUILD; }
    const real& cpu_tree_force  () const { return CPU_GRAV; }
    const real& cpu_ext_force   () const { return CPU_PEX; }
    const real  cpu_force       () const { return CPU_GRAV+CPU_PEX; }
    const real& cpu_density     () const { return CPU_DENS; }
    const real& cpu_softening   () const { return CPU_EPSI; }
    const real& cpu_longstep    () const { return CPU_STEP; }
    const real& cpu_total       () const { return CPU_TOTAL; }
    //--------------------------------------------------------------------------
    void  reset_cpu_total       (const real sec) const { CPU_TOTAL=sec; }
    //--------------------------------------------------------------------------
    void reset_softening(                          // resets softening params   
			 const soft_type,          // I: type of softening      
			 const unsigned,           // I: Nsoft                  
			 const real,               // I: eps                    
			 const kern_type);         // I: softening kernel       
    //--------------------------------------------------------------------------
    void reset_opening(const real th) const { falcON::reset_opening(th); }
    //--------------------------------------------------------------------------
    const real t_ini() const { return TINI; }
    //--------------------------------------------------------------------------
    basic_nbody(const bodies    *,                 // I: bodies                 
		const uint       ,                 // I: N_soft                 
		const real       ,                 // I: eps/eps_max            
		const real       ,                 // I: initial time           
		const soft_type  = global_fixed,   //[I: softening type]        
		const real       = half,           //[I: tolerance parameter]   
		const uint       = 0,              //[I: N_reuse]               
		const uint       = 6,              //[I: N_crit]                
		const kern_type  = p1,             //[I: softening kernel]      
		const uint       = 16u,            //[I: N_ref]                 
		const potential* = 0,              //[I: P_ex]                  
		const uint[4]    = nbdy::DIRECT    //[I: direct sum control]    
#ifdef __GRAPE
		,const bool      = false           //[I: use GRAPE?]            
#endif
		);
    //--------------------------------------------------------------------------
    virtual ~basic_nbody() {}
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::LeapFrogCode                                                   
  //////////////////////////////////////////////////////////////////////////////
  class LeapFrogCode :
    public    basic_nbody,
    protected LeapFrog
  {
    LeapFrogCode(const LeapFrogCode&);              // not implemented          
    LeapFrogCode& operator= (const LeapFrogCode&);  // not implemented          
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
  public:
    LeapFrogCode(const bodies    *,                // I: bodies                 
		 const uint       ,                // I: N_soft                 
		 const real       ,                // I: eps/eps_max            
		 const real       ,                // I: initial time           
		 const int        ,                // I: h0                     
		 const uint       = 0,             //[I: N_reuse]               
		 const vect     * = 0,             //[I: center of root]        
		 const soft_type  = global_fixed,  //[I: softening type]        
		 const real       = half,          //[I: tolerance parameter]   
		 const uint       = 6,             //[I: N_crit]                
		 const kern_type  = p1,            //[I: softening kernel]      
		 const uint       = 16u,           //[I: N_ref]                 
		 const potential* = 0,             //[I: P_ex]                  
		 const uint[4]    = nbdy::DIRECT   //[I: direct sum control]    
#ifdef __GRAPE
		 ,const bool      = false          //[I: use GRAPE?]            
#endif
		 );
    //--------------------------------------------------------------------------
    void full_step      (const vect* = 0);         //[I: center of root ]       
    const real& time() const { return LeapFrog::time(); }
    void long_stats      (std::ostream&) const;    // statistic of long_step    
    void short_stats     (std::ostream&) const;    // statistic of long_step    
    void short_stat_head (std::ostream&) const;    // table head for short_stats
    void short_stat_line (std::ostream&) const;    // line of proper size       
  };
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::BlockStepCode                                                  
  //////////////////////////////////////////////////////////////////////////////
  class BlockStepCode :
    public    basic_nbody,
    protected BlockStep
  {
    BlockStepCode(const BlockStepCode&);            // not implemented          
    BlockStepCode& operator= (const BlockStepCode&);// not implemented          
    //--------------------------------------------------------------------------
    // private methods                                                          
    //--------------------------------------------------------------------------
  private:
    int  W;                                        // width for output #levels  
    void set_sink_flags (const indx);
    void elementary_step(const indx, const vect*, const how_to_make_tree=free);
    void prepare(const int   ,                     // I: h0                     
		 const indx  ,                     // I: N_steps                
		 const vect* =0);                  //[I: center of root]        
    //--------------------------------------------------------------------------
    // public methods                                                           
    //--------------------------------------------------------------------------
  public:
    BlockStepCode(const bodies    *,               // I: bodies                 
		  const uint       ,               // I: N_soft                 
		  const real       ,               // I: eps/eps_max            
		  const real       ,               // I: initial time           
		  const int        ,               // I: h0                     
		  const indx       ,               // I: N_steps                
		  const uint       ,               // I: N_reuse                
		  const real       ,               // I: factor f_a: stepping   
		  const real       = 0,            //[I: factor f_p: stepping]  
		  const real       = 0,            //[I: factor f_c: stepping]  
		  const vect     * = 0,            //[I: center of root]        
		  const soft_type  = global_fixed, //[I: softening type]        
		  const real       = half,         //[I: tolerance parameter]   
		  const uint       = 6,            //[I: N_crit]                
		  const kern_type  = p1,           //[I: softening kernel]      
		  const uint       = 16u,          //[I: N_ref]                 
		  const potential* = 0,            //[I: P_ex]                  
		  const uint[4]    = nbdy::DIRECT  //[I: direct sum control]    
#ifdef __GRAPE
		  ,const bool      = false         //[I: use GRAPE?]            
#endif
		  );
    //--------------------------------------------------------------------------
    void reset_stepping (const real,               // I: f_a                    
			 const real = zero,        //[I: f_p]                   
			 const real = zero);       //[I: f_c]                   
    void full_step      (const vect* = 0);         //[I: center of root ]       
    const real& time    () const { return BlockStep::time(); }
    void dump_steps     (std::ostream& to) const { BlockStep::dump(to); }
    void long_stats     (std::ostream&) const;    // statistic of long_step    
    void short_stats    (std::ostream&) const;    // statistic of long_step    
    void short_stat_head(std::ostream&) const;    // statistic of long_step    
    void short_stat_line(std::ostream&) const;    // line of proper size       
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_nbdy_h
