// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// bcell.h                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// some macros for looping child nodes of a cell                               |
// class basic_cell                   fully inline                             |
// class basic_cell_access            fully inline                             |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_bcell_h
#define included_bcell_h 1

#ifndef included_flag_h
#  include <nbdy/flag.h>
#endif

#ifndef included_iomanip
#  include <iomanip>
#  define included_iomanip
#endif

using std::ostream;
using std::setw;
using std::setprecision;

////////////////////////////////////////////////////////////////////////////////
// useful macros                                                                
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------
// loop the cell children of a given cell                                       
//------------------------------------------------------------------------------
#define LoopCellKids(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
		     NAME)                         /* name for child cells   */\
  for(register CELL_TYPE::cell_child	           /* type of child cell     */\
      NAME  = CELL_PTER->begin_cell_children();    /* from first child       */\
      NAME != CELL_PTER->end_cell_children();      /* until beyond last      */\
    ++NAME)                                        // get next child            
//------------------------------------------------------------------------------
// loop the cell children of a given cell starting somewhere                    
//------------------------------------------------------------------------------
#define LoopCellSecd(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
                     CELL_START,                   /* start loop here        */\
		     NAME)                         /* name for child cells   */\
  for(register CELL_TYPE::cell_child	           /* type of child cell     */\
      NAME  = CELL_START;                          /* from start child       */\
      NAME != CELL_PTER->end_cell_children();      /* until beyond last      */\
    ++NAME)                                        // get next child            
//------------------------------------------------------------------------------
// loop the soul children of a given cell                                       
//------------------------------------------------------------------------------
#define LoopSoulKids(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
		     NAME)                         /* name for child souls   */\
  for(register CELL_TYPE::soul_child	           /* type of child cell     */\
      NAME  = CELL_PTER->begin_soul_children();    /* from first child       */\
      NAME != CELL_PTER->end_soul_children();      /* until beyond last      */\
    ++NAME)                                        // get next child            
//------------------------------------------------------------------------------
// loop the soul children of a given cell starting somewhere                    
//------------------------------------------------------------------------------
#define LoopSoulSecd(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
                     SOUL_START,                   /* start loop here        */\
		     NAME)                         /* name for child souls   */\
  for(register CELL_TYPE::soul_child	           /* type of child cell     */\
      NAME  = SOUL_START;                          /* from start child       */\
      NAME != CELL_PTER->end_soul_children();      /* until beyond last      */\
    ++NAME)                                        // get next child            
//------------------------------------------------------------------------------
// loop the soul descendants of a given cell                                    
//------------------------------------------------------------------------------
#define LoopAllSouls(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
		     NAME)                         /* name for child souls   */\
  for(register CELL_TYPE::soul_descendant          /* type of child cell     */\
      NAME  = CELL_PTER->begin_soul_descendants(); /* from first child       */\
      NAME != CELL_PTER->end_soul_descendants();   /* until beyond last      */\
    ++NAME)                                        // get next child            
//------------------------------------------------------------------------------
// loop the soul descendants of a given cell starting somewhere                 
//------------------------------------------------------------------------------
#define LoopSecSouls(CELL_TYPE,                    /* type of parent cell    */\
		     CELL_PTER,                    /* pointer to parent cell */\
                     SOUL_START,                   /* start loop here        */\
		     NAME)                         /* name for child souls   */\
  for(register CELL_TYPE::soul_descendant          /* type of child cell     */\
      NAME  = SOUL_START;                          /* from start child       */\
      NAME != CELL_PTER->end_soul_descendants();   /* until beyond last      */\
    ++NAME)                                        // get next child            
////////////////////////////////////////////////////////////////////////////////
// classes                                                                      
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  template<class CELL> class basic_cell_access;    // forward declaration       
  //////////////////////////////////////////////////////////////////////////////
  // nbdy::class basic_cell_data<CELL,SOUL>                                     
  //////////////////////////////////////////////////////////////////////////////
  template<class CELL, class SOUL>
  class basic_cell :  public basic_flag {
    //--------------------------------------------------------------------------
    // friendship                                                               
    //--------------------------------------------------------------------------
    friend class basic_cell_access<CELL>;          // allows data access        
    //--------------------------------------------------------------------------
    // member types                                                             
    //--------------------------------------------------------------------------
  public:
    typedef CELL                       cell_type;  // type of cell              
    typedef SOUL                       soul_type;  // type of soul              
    typedef cell_type                 *cell_pter;  // pointer to cell           
    typedef soul_type                 *soul_pter;  // pointer to soul           
    typedef const cell_type           *const_cell_pter;  // const pter to cell  
    typedef const soul_type           *const_soul_pter;  // const pter to soul  
    typedef typename SOUL::bodies_type bodies_type;// type of bodies            
    typedef real                       scalar_type;// type of real numbers      
    typedef vect                       vector_type;// type of vector of reals   
    typedef ten1                       evectr_type;// type of ten1 of reals     
    //--------------------------------------------------------------------------
    // data of class cell_basics (write access through class basic_cell_access) 
    //--------------------------------------------------------------------------
  private:
    scalar_type     RADIUS;                        // half edge length of cube  
    vector_type     CENTER;                        // center of cube            
    unsigned        NUMBER;                        // # soul descendants        
    soul_pter       CSOULS;                        // first soul child          
    cell_pter       CCELLS;                        // first cell child          
    short unsigned  NSOULS;                        // # soul children           
    short unsigned  NCELLS;                        // # cell children           
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  protected:
    basic_cell() {}
    //--------------------------------------------------------------------------
    // data access                                                              
    //--------------------------------------------------------------------------
  public:
    scalar_type    const&radius() const { return RADIUS; }
    vector_type    const&center() const { return CENTER; }
    unsigned       const&number() const { return NUMBER; }
    soul_pter      const&csouls() const { return CSOULS; }
    cell_pter      const&ccells() const { return CCELLS; }
    short unsigned const&nsouls() const { return NSOULS; }
    short unsigned const&ncells() const { return NCELLS; }
    //--------------------------------------------------------------------------
    // child access                                                             
    //--------------------------------------------------------------------------
    typedef cell_pter cell_child;
    cell_child begin_cell_children   () const { return CCELLS; }
    cell_child end_cell_children     () const { return CCELLS+NCELLS; }
    cell_child last_cell_child       () const { return CCELLS+NCELLS-1; }
    //--------------------------------------------------------------------------
    typedef soul_pter soul_child;
    soul_child begin_soul_children   () const { return CSOULS; }
    soul_child end_soul_children     () const { return CSOULS+NSOULS; }
    soul_child last_soul_child       () const { return CSOULS+NSOULS-1; }
    //--------------------------------------------------------------------------
    typedef soul_pter soul_descendant;
    soul_descendant begin_soul_descendants() const { return CSOULS; }
    soul_descendant end_soul_descendants  () const { return CSOULS+NUMBER; }
    soul_descendant last_soul_descendant  () const { return CSOULS+NUMBER-1; }
    //--------------------------------------------------------------------------
    // flag manipulations                                                       
    //--------------------------------------------------------------------------
    void reset_flag() {
      basic_flag::reset_flag();
      basic_flag::add_flag(basic_flag::AL_SINK);
    }
    //--------------------------------------------------------------------------
    void reset_sink_flag() {
      basic_flag::un_flag (basic_flag::SINK);
      basic_flag::add_flag(basic_flag::AL_SINK);
    }
    //--------------------------------------------------------------------------
    void add_sink_flag(const_soul_pter const&S) {
      basic_flag::add_part_flag(S,basic_flag::SINK);
      if(!is_sink(S)) basic_flag::un_flag(basic_flag::AL_SINK);
    }
    //--------------------------------------------------------------------------
    void add_sink_flag(const_cell_pter const &C) {
      basic_flag::add_part_flag(C,basic_flag::SINK);
      if(!al_sink(C)) basic_flag::un_flag(basic_flag::AL_SINK);
    }
    //--------------------------------------------------------------------------
    void add_body_flag(const_soul_pter const&S) {
      basic_flag::add_part_flag(S,BODY_FLAGS);
      if(!is_sink(S)) basic_flag::un_flag(AL_SINK);
    }
    //--------------------------------------------------------------------------
    void add_body_flag(const_cell_pter const &C) {
      basic_flag::add_part_flag(C,BODY_FLAGS);
      if(!al_sink(C)) basic_flag::un_flag(AL_SINK);
    }
    //--------------------------------------------------------------------------
    void update_sink_flag() {
      reset_sink_flag();
      LoopCellKids(typename cell_type,this,c) add_sink_flag(c);
      LoopSoulKids(typename cell_type,this,s) add_sink_flag(s);
    }
    //--------------------------------------------------------------------------
    // boolean information                                                      
    //--------------------------------------------------------------------------
    bool has_cell_children () const { return ncells()!=0; }
    bool has_soul_children () const { return nsouls()!=0; }
    bool is_twig           () const { return ncells()==0; }
    bool is_bough          () const { return nsouls()==0; }
    //--------------------------------------------------------------------------
    // dump cell data                                                           
    //--------------------------------------------------------------------------
    static void dump_head(ostream&o) {
      o<<" flg cells ncell souls nsoul number    radius            center";
    }
    //--------------------------------------------------------------------------
    void dump(ostream&o,
	      const_soul_pter const&S0,
	      const_cell_pter const&C0) const
    {
      o<<" "<<setw(3)<<flag_type(*this);
      if(ncells())
	o<<" "<<setw(5)<<int(CCELLS-C0);
      else
	o<<"     -";
      o<<" "<<setw(5)<<ncells()
       <<" "<<setw(5)<<int(CSOULS-S0)
       <<" "<<setw(5)<<nsouls()
       <<" "<<setw(6)<<number()
       <<" "<<setw(9)<<radius();
      for(register int i=0; i<NDIM; ++i)
	o<<" "<<setw(8)<<setprecision(4)<<center()[i];
    }
  };
  //////////////////////////////////////////////////////////////////////////////
  // nbdy::class basic_cell_access<CELL> (needed in tree building only)         
  //////////////////////////////////////////////////////////////////////////////
  template<class CELL>
  class basic_cell_access {
    //--------------------------------------------------------------------------
    // protected types and methods                                              
    //--------------------------------------------------------------------------
  protected:
    typedef CELL                       cell_type;  // type of associated cell   
    typedef typename CELL::soul_type   soul_type;  // type of associated soul   
    typedef typename CELL::soul_pter   soul_pter;  // pointer to associated soul
    typedef typename CELL::cell_pter   cell_pter;  // pointer to associated cell
    typedef basic_cell<CELL,soul_type> basic_cell; // type of basic cell        
    typedef basic_cell::scalar_type    scalar_type;// type of real numbers      
    typedef basic_cell::vector_type    vector_type;// type of vector of reals   
    //--------------------------------------------------------------------------
    static vector_type    &center_(basic_cell*C) { return C->CENTER; }
    static scalar_type    &radius_(basic_cell*C) { return C->RADIUS; }
    static unsigned       &number_(basic_cell*C) { return C->NUMBER; }
    static short unsigned &nsouls_(basic_cell*C) { return C->NSOULS; }
    static short unsigned &ncells_(basic_cell*C) { return C->NCELLS; }
    static soul_pter      &csouls_(basic_cell*C) { return C->CSOULS; }
    static cell_pter      &ccells_(basic_cell*C) { return C->CCELLS; }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_bcell_h
