// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// cell.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// defines                                                                     |
//                                                                             |
// class cell                         fully inline                             |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_cell_h
#define included_cell_h 1

#ifndef included_auxx_h
#  include <nbdy/auxx.h>
#endif

#ifndef included_bcell_h
#  include <nbdy/bcell.h>
#endif

#ifndef included_soul_h
#  include <nbdy/soul.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // nbdy::class cell                                                           
  //////////////////////////////////////////////////////////////////////////////
  class cell : public basic_cell<cell,soul> {
    cell           (const cell&);                  // not implemented           
    cell& operator=(const cell&);                  // not implemented           
    //--------------------------------------------------------------------------
    // friends                                                                  
    //--------------------------------------------------------------------------
    friend class gravity_tree;
    //--------------------------------------------------------------------------
    // member types                                                             
    //--------------------------------------------------------------------------
    typedef basic_cell<cell,soul> base;
    //--------------------------------------------------------------------------
    // static data and methods                                                  
    //--------------------------------------------------------------------------
#if NDIM==3
    static const int N_P3=6, N_P4=16;
    static const int N_EP=8, N_EI=9, N_EA=10, N_STUFF=11;
#else
    static const int N_P3=3, N_P4=7;
    static const int N_EP=7, N_EI=8, N_EA=9,  N_STUFF=10;
#endif
    static int N_eps  () { return N_EP; }
    static int N_stuff() { return N_STUFF; }
    //--------------------------------------------------------------------------
    // data of class cell                                                       
    //--------------------------------------------------------------------------
  private:
    scalar_type   *STUFF;                          // data for cell             
    scalar_type   *POLES;                          // multipole moments         
    scalar_type   *COEFFS;                         // Taylor coefficients       
    //--------------------------------------------------------------------------
    // constructor                                                              
    //--------------------------------------------------------------------------
  public:
    cell() : COEFFS(0) {}
    //--------------------------------------------------------------------------
    // simple manipulations                                                     
    //--------------------------------------------------------------------------
  public:
    void set_rcrit(const real ith) {
      STUFF[3] = ith * STUFF[2];
      STUFF[4] = util::square(STUFF[3]);
    }
    //--------------------------------------------------------------------------
    // non-const data access via members                                        
    //--------------------------------------------------------------------------
    evectr_type   cofm  () { return ten1(STUFF+5); }
    scalar_type  &mass  () { return STUFF[0]; }
    scalar_type  &imass () { return STUFF[1]; }
    scalar_type  &rmax  () { return STUFF[2]; }
    scalar_type  &eps   () { return STUFF[N_EP]; }
    scalar_type  &emin  () { return STUFF[N_EI]; }
    scalar_type  &emax  () { return STUFF[N_EA]; }
    scalar_type* &coeffs() { return COEFFS; }
    ten2          quad  () { return ten2(POLES); }
    ten3          octo  () { return ten3(POLES+N_P3); }
    ten4          hexa  () { return ten4(POLES+N_P4); }
    //--------------------------------------------------------------------------
    // const data access via friends                                            
    //--------------------------------------------------------------------------
#define CCCC const cell*const&C
    friend scalar_type const&radius(CCCC) { return C->radius(); }
    friend vector_type const&center(CCCC) { return C->center(); }
    friend unsigned    const&number(CCCC) { return C->number(); }
    friend evectr_type const cofm  (CCCC) { return ten1(C->STUFF+5); }
    friend scalar_type const&mass  (CCCC) { return C->STUFF[0]; }
    friend scalar_type const&imass (CCCC) { return C->STUFF[1]; }
    friend scalar_type const&rmax  (CCCC) { return C->STUFF[2]; }
    friend scalar_type const&rcrit (CCCC) { return C->STUFF[3]; }
    friend scalar_type const&rcrit2(CCCC) { return C->STUFF[4]; }
    friend scalar_type const&eps   (CCCC) { return C->STUFF[N_EP]; }
    friend scalar_type const&emin  (CCCC) { return C->STUFF[N_EI]; }
    friend scalar_type const&emax  (CCCC) { return C->STUFF[N_EA]; }
    friend scalar_type*const&coeffs(CCCC) { return C->COEFFS; }
    friend ten2              quad  (CCCC) { return ten2(C->POLES); }
    friend ten3              octo  (CCCC) { return ten3(C->POLES+N_P3); }
    friend ten4              hexa  (CCCC) { return ten4(C->POLES+N_P4); }
    //--------------------------------------------------------------------------
    // boolean information via friends                                          
    //--------------------------------------------------------------------------
    friend bool is_source        (CCCC) { return nbdy::mass(C)!=zero; }
    friend bool has_cell_children(CCCC) { return C->has_cell_children(); }
    friend bool has_soul_children(CCCC) { return C->has_soul_children(); }
    friend bool is_twig          (CCCC) { return C->is_twig(); }
    friend bool is_bough         (CCCC) { return C->is_bough(); }
    //--------------------------------------------------------------------------
    // other const methods and friends                                          
    //--------------------------------------------------------------------------
    friend real xmin(CCCC) { return nbdy::cofm(C).min() - nbdy::rmax(C); }
    friend real xmax(CCCC) { return nbdy::cofm(C).max() + nbdy::rmax(C); }
#undef CCCC
    //--------------------------------------------------------------------------
    // dump cell data                                                           
    //--------------------------------------------------------------------------
    static void dump_head(std::ostream&o) {
      o<<"     #";
      base::dump_head(o);
      o<<"              mass              cofm                  rmax";
    }
    //--------------------------------------------------------------------------
    void dump(std::ostream& o,
	      const_soul_pter const&S0,
	      const_cell_pter const&C0) const {
      o<<" "<<std::setw(5)<<int(this-C0);
      base::dump(o,S0,C0);
      o<<" "<<setw(8)<<nbdy::mass(this);
      for(register int i=0; i<NDIM; i++)
	o<<" "<<setw(8)<<setprecision(4)<<nbdy::cofm(this)[i];
      o<<" "<<setw(12)<<nbdy::rmax(this);
    }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_cell_h
