// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// grav.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// versions/changes (documented since June 2002)                               |
//                                                                             |
// 11/06/2002 - added option for full potential support (FULL_OPT):            |
//              if specified, the octopole contributes to the potential        |
//            - support for kernels Fn and Kn withdrawn                        |
//                                                                             |
//-----------------------------------------------------------------------------+
#ifndef included_grav_h
#define included_grav_h

#ifndef included_kern_h
#  include <nbdy/kern.h>
#endif

#ifndef included_cell_h
#  include <nbdy/cell.h>
#endif

#ifndef included_tayl_h
#  include <nbdy/tayl.h>
#endif

#ifndef included_stat_h
#  include <nbdy/stat.h>
#endif

#ifndef included_drct_h
#  include <nbdy/drct.h>
#endif
////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
#ifdef __GRAPE
  class grape_arrays;                                // forward declaration     
#endif
  //////////////////////////////////////////////////////////////////////////////
  //                                                                            
  // class nbdy::gravity_kernel                                                 
  //                                                                            
  // This class is at the heart of the algorithm. It serves as INTERACTOR in    
  // the template function MutualInteract(), defined in iact.h, which encodes   
  // the interaction phase in a most general way.                               
  // class nbdy::gravity_kernel has member functions for soul-soul, soul-cell,  
  // cell-cell, and cell-self interactions (methods interact() below), as well  
  // as a function for the evaluation phase, method evaluate_grav().            
  //                                                                            
  // In order to allow for more efficient usage of free memory, the cells'      
  // pointer to the memory containing the Taylor coefficients is initialized    
  // only when needed, i.e. when the corresponding cell actually has non-zero   
  // Taylor coefficients. Reversely, when the memory is no longer needed, i.e.  
  // when the coefficients have been accumulated in the evaluation phase, the   
  // memory is taken back into the pool.                                        
  //                                                                            
  //////////////////////////////////////////////////////////////////////////////
  class gravity_kernel : public TaylorExpansion {
  private:
    //--------------------------------------------------------------------------
    // types                                                                    
    //--------------------------------------------------------------------------
    typedef cell cell_type;                            // type of cell and soul 
    typedef soul soul_type;                            // that we are dealing w.
    //--------------------------------------------------------------------------
    enum who {                                         // what subsouls are sink
      none = 0,                                        //   none                
      some = 1,                                        //   some, but not all   
      all  = 2                                         //   all                 
    };
    //--------------------------------------------------------------------------
    // data                                                                     
    //--------------------------------------------------------------------------
    bool                  FULL_POT;                    // Pth pole in pot       
    uint                  N_PRE[3], N_POST[3];         // direct sums control   
#ifdef __GRAPE
    mutable grape_arrays *GARR;                        // for usage with GRAPE  
#endif
    soft_type             SOFT;                        // global/individual     
    kern_type             KERN;                        // softening kernel      
    TreeStats            *STAT;                        // statistics            
    mutable real          EPS, EPQ, IEP, IEQ;          // eps & powers of eps   
    //--------------------------------------------------------------------------
    // private methods                                                          
    //--------------------------------------------------------------------------
    who  who_is_sink      (const cell*) const;       
    void many_global      (const bool,soul*,const who,soul*,const soul*) const;
    void many_individual  (const bool,soul*,const who,soul*,const soul*) const;
    void many_global      (cell*, soul*) const;
    void many_individual  (cell*, soul*) const;
    void many_global      (cell*, cell*) const;
    void many_individual  (cell*, cell*) const;
    void many_global      (cell*) const;
    void many_individual  (cell*) const;
    //--------------------------------------------------------------------------
    template<bool F> void grav(vect&, const real, cell*, soul*) const;
    template<bool F> void grav(vect&, const real, cell*, cell*) const;
    //--------------------------------------------------------------------------
    template<bool F> bool mutual_global    (cell*, soul*) const;
    template<bool F> bool mutual_individual(cell*, soul*) const;
    template<bool F> bool mutual_global    (cell*, cell*) const;
    template<bool F> bool mutual_individual(cell*, cell*) const;
    //--------------------------------------------------------------------------
    // non-const public methods                                                 
    //--------------------------------------------------------------------------
  public:
    gravity_kernel(      TreeStats*,                  // I: statistics          
		   const kern_type ,                  // I: type of kernel      
		   const uint      ,                  // I: initial pool size   
		   const real     = zero,             //[I: softening length]   
		   const soft_type= global,           //[I: type of softening]  
		   const uint[4]  = DIRECT,           //[I: direct sum control] 
		   const bool     = false,            //[I: use Pth pole in pot]
		   const uint     = 3                 //[I: expansion order]
#ifdef __GRAPE
		  ,const uint     = 0                 //[I: GRAPE?: # souls]    
		  ,const soul*    = 0                 //[I: GRAPE?: souls]      
#endif
			              );
    //--------------------------------------------------------------------------
    ~gravity_kernel();                                // destructor             
    //--------------------------------------------------------------------------
    void set_softening(const soft_type ST) {          // set softening type     
      SOFT=ST;                                        //   to new value         
    }
    //--------------------------------------------------------------------------
    // const public methods                                                     
    //--------------------------------------------------------------------------
#ifdef __GRAPE
    bool       using_grape  ()     const { return GARR != 0; }
#endif
    //--------------------------------------------------------------------------
    const real&current_eps  ()     const { return EPS; }
    const real&current_epsq ()     const { return EPQ; }
    //--------------------------------------------------------------------------
    bool interact(cell*) const;                       // cell-self interaction  
    bool interact(cell*, cell*) const;                // cell-cell interaction  
    bool interact(cell*, soul*) const;                // cell-body interaction  
    void interact(soul*, soul*) const;                // body-body interaction  
    void direct_summation(cell*) const;               // direct summation       
    //--------------------------------------------------------------------------
    void evaluate_grav(cell*C) const {                // evaluation phase       
#ifdef __GRAPE
      TaylorExpansion::evaluate_grav(C,GARR);
#else
      TaylorExpansion::evaluate_grav(C);
#endif
    }
    //--------------------------------------------------------------------------
    bool split_first(const cell*A, const cell*B) const {
      return is_twig(B) || (!is_twig(A) && rmax(A) > rmax(B));
    }
    //--------------------------------------------------------------------------
    real greens_function(vect&) const;
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_grav_h
