// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// tcdm.h                                                                      |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2002                                               |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
//                                                                             |
// class for a self-consistent spherical stellar dynamical model with radial   |
// density profile                                                             |
//                                                                             |
//                  sech(r/b)                                                  |
// rho(r) = C ---------------------                                            |
//            r^alfa (r+a)^(3-alfa)                                            |
//                                                                             |
// where:                                                                      |
//                                                                             |
// alfa = inner density exponent,                                              |
// a    = scale- or break-radius,                                              |
// b    = truncation radius, and                                               |
// C    = normalization constant                                               |
//                                                                             |
// C is set such that the maximum circular velocity equals a given value.      |
//                                                                             |
//-----------------------------------------------------------------------------+

#ifndef included_tcdm_h
#define included_tcdm_h

////////////////////////////////////////////////////////////////////////////////
namespace nbdy {
  //////////////////////////////////////////////////////////////////////////////
  // class nbdy::tcdm_model                                                     
  //////////////////////////////////////////////////////////////////////////////
  class tcdm_model {
    //--------------------------------------------------------------------------
    // private data                                                             
    //--------------------------------------------------------------------------
  private:
    double   A,c,a,b;                             // parameters                 
    int      n,n1,nm;                             // size of tables             
    double   *r,*lr,*m,*ps,*sq,*lf;               // tables                     
    double   ps0,sq0,vqm,etot;
    double   ab,ca,cac,fac,ln_a,ln_fac;
    //--------------------------------------------------------------------------
    // private methods                                                          
    //--------------------------------------------------------------------------
    void setup(const double);                     // I: max v_circ              
    void setfE();
    double fiso(const double&) const;             // f(Eps=-E)                  
    double psix(const double&) const;             // Psi(x=r/a)                 
    double rhox(const double&) const;             // rho(x=r/a)                 
    double cumx(const double&) const;             // M(<x=r/a)                  
    //--------------------------------------------------------------------------
    // construction & destruction                                               
    //--------------------------------------------------------------------------
  public:
    tcdm_model(const double = 1,                  //[I: inner density exponent  
	       const double = 1,                  //[I: break radius a]         
	       const double = 10,                 //[I: truncation radius b]    
	       const double = 1);                 //[I: maximal v_circ]         
    //--------------------------------------------------------------------------
    ~tcdm_model();                                // free memory                
    //--------------------------------------------------------------------------
    // non-const methods                                                        
    //--------------------------------------------------------------------------
    void set_vc_max (const double&);              // set normalization such that
                                                  // the maximal v_circ matches 
    //--------------------------------------------------------------------------
    // const methods: global properties                                         
    //--------------------------------------------------------------------------
    double const&inner_slope () const { return A; }
    double const&break_radius() const { return a; }
    double const&trunc_radius() const { return b; }
    double const&mass_normal () const { return c; }
    double       central_pot () const { return -ca *ps0; }
    double       total_mass  () const { return c   *m[n1]; }
    double       total_energy() const { return c*ca*etot; }
    double       vcirc_square() const { return ca*vqm; }
    //--------------------------------------------------------------------------
    // const methods: local properties                                          
    //--------------------------------------------------------------------------
    double rho      (const double&) const;        // rho(r)                     
    double pot      (const double&) const;        // Phi(r)                     
    double cum      (const double&) const;        // M(<r)                      
    double vcq      (const double&) const;        // vc^2(r)                    
    double siq      (const double&) const;        // isotropic sigma^2(r)       
    double f_E      (const double&) const;        // isotropic f(E)             
    //--------------------------------------------------------------------------
    // phase-space sampling                                                     
    //--------------------------------------------------------------------------
    double random   (                             // R: potential at r          
		     double(*)(),                 // I: random number generator 
		     double&,                     // O: radius                  
		     double&) const;              // O: velocity                
    //--------------------------------------------------------------------------
    // direct const access to grids                                             
    //--------------------------------------------------------------------------
    const int& N_grid()            const { return n; }
    double rad_grid (const int &i) const { return a  * r[i]; }
    double lnr_grid (const int &i) const { return ln_a + lr[i]; }
    double cum_grid (const int &i) const { return c  * m[i]; }
    double psi_grid (const int &i) const { return ca * ps[i]; }
    double vcq_grid (const int &i) const { return ca * m[i] / r[i]; }
    double siq_grid (const int &i) const { return ca * sq[i]; }
    double lnf_grid (const int &i) const { return ln_fac + lf[i]; }
  };
}
////////////////////////////////////////////////////////////////////////////////
#endif // included_tcdm_h
