////////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <ctime>
#include <cstdlib>
#ifdef unix
#  define GIVE_HOST_INFO
  extern "C" {
#  include <unistd.h>
#  include <pwd.h>
  }
#endif
#include <king.h>
#include <body.h>
#include <util/Pi.h>
#include <util/rand.h>

using namespace nbdy;
using namespace util;
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
  const double Msun = 4.498658966346282e-6;        // solar mass in N-body units
  const double kms  = 1./0.977775320024919;        // km/s in N-body units      

  if(argc!=7 && argc!=13) {
    cerr<<
      " \"King W0 N S M R Sym [x,y,z, vx,vy,vz]\"\n"
      " writes initial conditions from King model in YANC format to stdout\n"
      "  W0:               dimensionless central potential\n"
      "  N:                number of positions & velocities\n"
      "  S = long:         seed for RNG if S=0, use sobol\n"
      "  M:                if > 0: mass in solar units\n"
      "                    if = 0: use natural rather than N-body units\n"
      "  R:                tidal/core (if<0) radius in kpc\n"
      "  C = 0/1:          ensure centroid?\n"
      "  [x,y,z,vx,vy,vz]: optional: position of origin [kpc,km/s]\n"
      " units imply G=1 and\n"
      "  either imply M=1, Rc/Rt=1 (if M=0 given)\n"
      "  or     are kpc and Myr (if M is given in M_sun)\n";
    exit(1);
  }

  bool           shift=false,natural=false;
  int            p=1, N, Sy;
  double         Mt,mass,Rt,W0;
  vector<double> X0=0., V0=0.;
  long           S;
  //----------------------------------------------------------------------------
  // 1. read parameters                                                         
  //----------------------------------------------------------------------------
  W0 = atof(argv[p++]);
  N  = atoi(argv[p++]);
  S  = atoi(argv[p++]);
  Mt = atof(argv[p++]);
  if(Mt)
    mass    = Mt * Msun;
  else   {
    natural = true;
    mass    = 1.;
  }
  Rt = atof(argv[p++]);
  Sy = atoi(argv[p++]);
  if(argc==13) {
    shift = true;
    for(register int i=0; i!=3; ++i) X0[i] = atof(argv[p++]);
    for(register int i=0; i!=3; ++i) V0[i] = atof(argv[p++]);
    if(!natural) V0 *= kms;
  }
  //----------------------------------------------------------------------------
  // 2. create King model & write header                                        
  //----------------------------------------------------------------------------
  king_model KM(W0);
  if(Rt>0) KM.reset_scales_tidal(mass, Rt);
  else     KM.reset_scales_core (mass,-Rt);
#ifdef GIVE_HOST_INFO
  time_t now = time(0);
  char host[100]; gethostname(host,100);
#endif
  cout<<
    "## -----------------------------------------------------------------------\n"
    "## King model initial conditions in YANC format\n"
    "## created\n"
#ifdef GIVE_HOST_INFO
    "## on              "<<ctime(&now)<<
    "## with program    \"King\" of Walter Dehnen (MPIA Heidelberg)\n"
    "## by user         \""<<(getpwuid(geteuid())->pw_name)<<"\"\n"
    "## on machine      \""<<host<<"\"\n##\n"
#else
    "## with program    \"King\" of Walter Dehnen (MPIA Heidelberg)\n##\n"
#endif
    "## King W0:        "<<W0<<"\n";
  if(natural) {
    if(Rt>0) cout<<
	       "## in natural units (G=M=Rt=1)\n"
	       "## core  radius:   "<<KM.core_radius() <<"\n";
    else     cout<<
	       "## in natural units (G=M=Rc=1)\n"
	       "## tidal radius:   "<<KM.tidal_radius()<<"\n";
  } else {
    cout<<
      "## in N-body units (G=1,[R]=kpc,[T]=Myr)\n"
      "## total mass:     "<<mass<<" = "<<Mt<<" Msun\n"
      "## core  radius:   "<<KM.core_radius() <<" kpc\n"
      "## tidal radius:   "<<KM.tidal_radius()<<" kpc\n";
  }
  if(shift) cout<<"## shifted to pos: "<<X0<<"\n"
		<<"## shifted to vel: "<<V0<<"\n";
  if(S) cout<<"## using pseudo-random numbers\n";
  else  cout<<"## using quasi-random numbers\n";
  cout<<
    "## -----------------------------------------------------------------------\n"
    "# time           0\n"
    "# N              "<<N<<"\n"
    "# data_format    0\n"
    "# start_data\n";
  //----------------------------------------------------------------------------
  // 3. create initial (x,v)                                                    
  //----------------------------------------------------------------------------
  bodies BB(N);
  const    double m = mass/double(N);
  register double rad,vel,cth,R,phi;
  Sobol           s1,s2,s3,s4,s5,s6;
  Random3         R3(S);
  LoopBodies(nbdy::bodies,(&BB),Bi) {
    //                                                                          
    // 3.1 set mass and get r and v                                             
    //                                                                          
    Bi.mass() = m;
    if(S) KM.random(s1(),s4(), rad,vel);
    else  KM.random(R3(),R3(), rad,vel);
    //                                                                          
    // 3.2 get x,y,z                                                            
    //                                                                          
    cth       = S? 2*R3()-1 : 2*s2()-1;
    R         = rad * sqrt(1.-cth*cth);
    phi       = TPi*( S? R3() : s3() );
    Bi.pos(0) = R * cos(phi);
    Bi.pos(1) = R * sin(phi);
    Bi.pos(2) = rad * cth;
    //                                                                          
    // 3.3 get vx,vy,vz                                                         
    //                                                                          
    cth       = S? 2*R3()-1 : 2*s5()-1;
    R         = vel * sqrt(1.-cth*cth);
    phi       = TPi*( S? R3() : s6() );
    Bi.vel(0) = R * cos(phi);
    Bi.vel(1) = R * sin(phi);
    Bi.vel(2) = vel * cth;
  }
  //----------------------------------------------------------------------------
  // 4. centrate position and velocity on origin or (x0,v0)                     
  //----------------------------------------------------------------------------
  register vector<double> cpos(0.), cvel(0.);
  if(Sy) {
    LoopBodies(nbdy::bodies,(&BB),Bi) {
      cpos += nbdy::pos(Bi);
      cvel += nbdy::vel(Bi);
    }
    cpos/= double(N);
    cvel/= double(N);
  }
  cpos-= X0;
  cvel-= V0;
  LoopBodies(nbdy::bodies,(&BB),Bi) {
    Bi.pos() -= cpos;
    Bi.vel() -= cvel;
  }
  //----------------------------------------------------------------------------
  // 5. output of (x,v)                                                         
  //----------------------------------------------------------------------------
  BB.write_yanc_ascii(cout,io::mxv);
  cout.flush();
}
