// -*- C++ -*- std header                                                      |
//-----------------------------------------------------------------------------+
//                                                                             |
// TestGrav.cc                                                                 |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2001                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <iostream>
#include <iomanip>
#include <ctime>
#include <falcON.h>
#include <body.h>

using std::cerr;
using std::cout;
using std::setprecision;
using namespace util;
using namespace nbdy;

const double Pi         = 3.14159265358979323846264338328;
const double Pih        = 0.5 * Pi;
const double TPi        = 2.0 * Pi;

double ms_Force_gamma(const double gamma)
{
  register double x=5-3*gamma,f;
  if(x < 0.) return 0.;
  f =1/x;
  f-=4/(x+=1);
  f+=6/(x+=1);
  f-=4/(x+=1);
  f+=1/(x+=1);
  return (3-gamma)*f;
}

double ms_density_gamma(const double gamma)
{
  register double x=3-3*gamma,f;
  if(x < 0.) return 0.;
  f = 1/x;
  f-= 8/(x+=1);
  f+=28/(x+=1);
  f-=56/(x+=1);
  f+=70/(x+=1);
  f-=56/(x+=1);
  f+=28/(x+=1);
  f-= 8/(x+=1);
  f+= 1/(x+=1);
  return cube(3-gamma)*f/square(4*Pi);
}

int main(int argc, char* argv[])
{
#ifdef TWODIMENSIONAL
  if(argc < 8 || argc > 15) {
    cerr<<
      " \"TestGrav MOD N S THE C K EPS [Nc] [s] [Ncb0] [Ncb1] [Ncc] [Ncs] [T]\" with \n"
      " MOD = 3/4           : Kuzmin disk / homogeneous disk\n"
#else
#ifdef __GRAPE
  if(argc < 9 || argc > 17) {
    cerr<<
      " \"TestGrav MOD gamma N S THE C K EPS [Nc] [s] [Ncb0] [Ncb1] [Ncc] [Ncs] [G] [T]\" with \n"
#else
  if(argc < 9 || argc > 16) {
    cerr<<
      " \"TestGrav MOD gamma N S THE C K EPS [Nc] [s] [Ncb0] [Ncb1] [Ncc] [Ncs] [T]\" with \n"
#endif
      " MOD = 0/1/2/3/4     : hom. sphere / Plummer / gamma-model"
      " / Kuzmin disk / hom. disk\n"
      " gamma               : parameter of model (if any)\n"
#endif
      " N                   : number of positions\n"
      " S = long            : seed for RNG\n"
      " THE                 : opening angle, if THE<0, use theta=const\n"
      " C = 0/1             : center root on origin or not ?\n"
      " K = n               : P_n softening kernel\n"
      " EPS                 : max / fixed softening length\n"
      " Nc   (default 6)    : don't split cell with <= Nc bodies\n"
      " s    (default 1)    : interweave interaction & evaluation phase?\n"
      " Ncb0 (default 3)    : N_cb^pre\n"
      " Ncb1 (default 128)  : N_cb^post\n"
      " Ncc  (default 5)    : N_cc^post\n"
      " Ncs  (default 64)   : N_cs\n"
#ifndef TWODIMENSIONAL
#ifdef __GRAPE
      " G    (default 0) 0/1: use GRAPE\n"
#endif
#endif
      " T    (default 0) 0/1: dump nodes to tree.cells and tree.souls\n";
    exit(1);
  }

  register clock_t  cpu0 = clock(), cpu1;
  register real     time_setup, time_grow, time_esti;
  const real        rmax = 1.e3;
  int               p=1, MOD, C, KER, T=0;
#ifdef __GRAPE
  int               G=0;
#endif
  kern_type         K;
  soft_type         Soft;
  long              Seed;
  unsigned          N, Ncrit=6, split=1, NCB0=3,NCB1=128,NCC=5,NCS=64;
  real              EPS, TH;

  MOD  = atoi(argv[p++]);
#ifndef TWODIMENSIONAL
  real GAM;
  GAM  = atof(argv[p++]);
#endif
  N    = atoi(argv[p++]);
  Seed = atoi(argv[p++]);
  TH   = atof(argv[p++]);
  C    = atoi(argv[p++]);
  KER  = atoi(argv[p++]);
  EPS  = atof(argv[p++]);
  if(argc>p) Ncrit = atoi(argv[p++]);
  if(argc>p) split = atoi(argv[p++]);
  if(argc>p) NCB0  = atoi(argv[p++]);
  if(argc>p) NCB1  = atoi(argv[p++]);
  if(argc>p) NCC   = atoi(argv[p++]);
  if(argc>p) NCS   = atoi(argv[p++]);
  unsigned DIR[4]; DIR[0]=NCB0; DIR[1]=NCB1; DIR[2]=NCC; DIR[3]=NCS;
#ifndef TWODIMENSIONAL
#ifdef __GRAPE
  if(argc>p) G     = atoi(argv[p++]);
#endif
#endif
  if(argc>p) T     = atoi(argv[p++]);
  bodies *BODIES = new bodies(N);
  vect  *A = new vect[N];
  real *RH = new real[N];
  srand48(Seed);
  K    = kern(KER);
  Soft = nbdy::global;
  EPS  = abs(EPS);

#ifndef TWODIMENSIONAL
  const    real g1=1.-GAM, g2=2.-GAM, ig2=1./g2, g3=3-GAM, ig3=1./g3;
#endif
  register real tm,r=zero,R,phi,fr,P,rh=zero;
#ifndef TWODIMENSIONAL
  register real cth;
#endif
  register double mi = one/real(N);
  vect origin=zero, *porigin = C? &origin : 0;
  LoopBodies(bodies,BODIES,Bi) {
    do {
      switch(MOD) {
#ifndef TWODIMENSIONAL
      case 0:  // homogeneous
	r  = pow(drand48(),1./3.);
	P  = 0.5*(3*r*r-1);
	fr =-1.;
	rh = 0.75/Pi;
	break;
      case 1:  // Plummer
	tm = pow(drand48(),2./3.);
	tm/= 1.-tm;                    // r^2
	r  = sqrt(tm);
	P  =-1./sqrt(tm+1);
	fr = P*P*P;
	rh = 0.75/Pi*pow(-P,5u);
	break;
      case 2:  // gamma-model
	tm = pow(drand48(),double(ig3));
	r  = tm / (1.-tm);
	P  = (GAM==2.)?  log(tm) : -ig2*(1-POW(tm,g2));
	fr =-POW(tm,g1) * square(1-tm) / r;
	rh = 0.25*g3/Pi * pow(1.-tm,4u)/POW(tm,GAM);
	break;
#endif
      case 3:
	tm = drand48();
	r  = sqrt((2-tm)*tm)/(1.-tm);
	P  =-1./sqrt(r*r+1);
	fr = P*P*P;
	rh =-0.5/Pi*fr;
	break;
      case 4:
	tm = drand48();
	r  = sqrt(tm);
	P  = 2*(r*r/3.-2);
	fr =-4*r/3.;
	rh = 1.0/Pi;
	break;
       default:
	cerr<<" unknown model\n";
	exit(1);
      }
    } while(r>rmax);
#ifdef TWODIMENSIONAL
    R = r;
#else
    cth = (MOD==3) ? zero : 2*drand48()-1;
    R   = r*sqrt(1.-cth*cth);
    Bi.pos(2) = r * cth;
#endif
    phi       = TPi*drand48();
    Bi.pos(0) = R * sin(phi);
    Bi.pos(1) = R * cos(phi);
    Bi.mass() = mi;
    Bi.flag_as_sink();
    A [index(Bi)] = fr * pos(Bi);   // exact force
    RH[index(Bi)] = rh;             // exact density
  }
//   // create file with masses & positions
//   ofstream out("positions.dat");
//   out<<N<<"\n";
//   for(Bi=B,i=0; i!=N; ++Bi,++i)
//     out<<setprecision(12)<<mi<<" "<<pos(Bi)<<"\n";
//   out.close();
//   //                                    
  cpu1 = clock();
  time_setup = (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
  cpu0 = cpu1;
  cout<<"\n"<<setprecision(6)
      <<" time needed for set up of X_i:           "<<time_setup<<"\n";
  falcON falcon(BODIES,EPS,TH,K);
  falcon.grow(Ncrit,porigin);
  cpu1 = clock();
  time_grow = (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
  cpu0 = cpu1;
  cout<<setprecision(6)
      <<" time needed for falcON::grow():          "<<time_grow<<"\n";
  if(T) falcon.dump_nodes("tree.cells","tree.souls");

  cpu0 = clock();
#ifdef __GRAPE
  falcon.approximate_gravity(split,DIR,G);
#else
  falcon.approximate_gravity(split,DIR);
#endif
  cpu1 = clock();
  time_esti = (cpu1 - cpu0) / real(CLOCKS_PER_SEC);
  cout<<setprecision(6)
      <<" time needed for falcON::approximate():   "<<time_esti<<"\n";
  falcon.stats(cout);
  cout<<"\n";

  register real ase=0.,damx=0.,dacc;
  register vect SF=zero;
  LoopBodies(bodies,BODIES,Bi) {
    dacc = norm(A[index(Bi)] - acc(Bi));
    ase += dacc;
    damx = max(damx,dacc);
    SF  += mass(Bi)*acc(Bi);
//     cerr<<setfill(' ')
// 	<<setw(13)<<setprecision(8)<<Bi.pos()[0]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.pos()[1]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.pos()[2]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.acc()[0]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.acc()[1]<<" "
// 	<<setw(13)<<setprecision(8)<<Bi.acc()[2]<<endl;
  }
  ase /= double(N);
  switch(MOD){
  case 1: ase *= 13.125; break;
  case 3: ase *= 17.5;   break;
  case 4: ase *= 1.125;  break;
  case 0: ase /= 0.6;    break;
#ifndef TWODIMENSIONAL
  case 2: if(GAM<1.66) ase /= ms_Force_gamma(GAM); break;    
#endif      
  default: break;
  }
#ifndef TWODIMENSIONAL
  if(MOD==2 && GAM>=1.66)
    cout<<setprecision(10)<<" ASE(F)           = "<< ase <<"\n";
  else
#endif      
    cout<<setprecision(10)<<" ASE(F)/<F^2>     = "<< ase <<"\n";
  cout<<" max (dF)^2       = "<< damx<<"\n"
      <<" Sum m_i F_i      = "<< SF  <<"\n";
//  cout<<endl;
//   for(;;) {
//     cerr<<"i = "; cin>>i;
//     cerr<<" "<<acc(B+i)<<" "<<neighbours(B+i)<<"\n";
//   }
  delete BODIES;
}
