//-----------------------------------------------------------------------------+
//                                                                             |
// nbdy.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 2000-2002                                          |
// e-mail:   dehnen@mpia.de                                                    |
// address:  Max-Planck Institut fuer Astronomie,                              |
//           Koenigstuhl 17, D-69117 Heidelberg, Germany                       |
//                                                                             |
//-----------------------------------------------------------------------------+
#include <nbdy.h>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <util/inln.h>
#include <util/Pi.h>
using std::ostream;
using std::endl;
using std::ios;
using std::setw;
using std::setprecision;
using namespace util;
using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
// useful macro looping bodies                                                  
////////////////////////////////////////////////////////////////////////////////
#define LoopMyBodies LoopBodies(bodies,BODIES,Bi)
////////////////////////////////////////////////////////////////////////////////
// auxiliary inline function                                                    
////////////////////////////////////////////////////////////////////////////////
inline void record_cpu(clock_t& c0, real& CPU) {
  register clock_t c1 = clock();
  CPU += (c1-c0)/real(CLOCKS_PER_SEC);
  c0 = c1;
  }
////////////////////////////////////////////////////////////////////////////////
// class fasstree::basic_nbody                                                  
////////////////////////////////////////////////////////////////////////////////
inline void basic_nbody::reset_cpus() const
{
  CPU_BUILD   = zero;
  CPU_GRAV    = zero;
  CPU_PEX     = zero;
  CPU_DENS    = zero;
  CPU_EPSI    = zero;
  CPU_STEP    = zero;
}
//------------------------------------------------------------------------------
inline void basic_nbody::update_cpu_total() const
{
  register clock_t C_NEW = clock();
  CPU_TOTAL += (C_NEW-C_OLD)/real(CLOCKS_PER_SEC);
  C_OLD = C_NEW;
}
//------------------------------------------------------------------------------
inline void basic_nbody::set_tree(const vect *x, const how_to_make_tree how)
{
  register clock_t cpu = clock();                  // get current CPU time      
  if(how == grow ||                                // IF must falcON::grow()    
     (how == free && N_REUSED >= NREUSE) ) {       // OR may and need to     >  
    falcON::grow(NCRIT,x);                         //   grow() new tree         
    N_REUSED=0u;                                   //   reset reuse() counter   
  } else {                                         // < ELSE >               >  
    falcON::reuse();                               //   reuse() old tree        
    ++N_REUSED;                                    //   count falcON::reuse()s  
  }                                                // <                         
  record_cpu(cpu,CPU_BUILD);                       // record CPU time required  
}
//------------------------------------------------------------------------------
void basic_nbody::set_forces()
{
  register clock_t cpu = clock();
  falcON::approximate_gravity(true,DIRECT          //  compute sink forces      
#ifdef __GRAPE
			      ,GRAPE
#endif
			      );
  record_cpu(cpu,CPU_GRAV);                        //  record CPU               
  if(PEX) {                                        //  if external potential    
    register vect A;
    LoopMyBodies if(is_sink(Bi)) {                 //    loop body sinks        
      Bi.pex()  = PEX->pot_f(A,pos(Bi),time());    //      set external pot     
      Bi.acc() += A;                               //      add external acc     
    }
    record_cpu(cpu,CPU_PEX);                       //  record CPU               
  }
}
//------------------------------------------------------------------------------
void basic_nbody::reset_softening(const soft_type sf,
				  const uint ns, const real e,
				  const kern_type ker)
{
  SOFTENING = sf;
  NSOFT     = ns;
  switch(SOFTENING) {
  case individual_adaptive:
  case individual_fixed:
    EMAX = e;
    falcON::reset_softening(-one,ker);
    break;
  default:
    EMAX = zero;
    falcON::reset_softening(e,ker);
  }
}
//------------------------------------------------------------------------------
void basic_nbody::set_softening_lengths()
{
  // set eps_i for all bodies such that there are ~NSOFT bodies in the softening
  // volume, as estimated from the local number-density estimates.              
#if NDIM==3
  const real fac = 0.75*NSOFT/Pi;
#else
  const real fac = NSOFT/Pi;
#endif	
  register clock_t  cpu = clock();                 // cpu time                  
  LoopMyBodies Bi.flag_as_sink();                  // flag every body           
  estimate_n(NSOFT);                               // estimate n_i              
  LoopMyBodies                                     // loop all bodies           
#if NDIM==3
    Bi.eps() = min(EMAX,POW(fac/rho(Bi),third));   //   set eps_i (3D)          
#else
    Bi.eps() = min(EMAX,SQRT(fac/rho(Bi)));        //   set eps_i (2D)          
#endif	
  record_cpu(cpu,CPU_EPSI);                        // record CPU                
}
//------------------------------------------------------------------------------
basic_nbody::basic_nbody(const bodies   *b,        // I: bodies                 
			 const uint      ns,       // I: N_soft                 
			 const real      e,        // I: eps/eps_max            
			 const real      ti,       // I: t_initial              
			 const soft_type sf,       //[I: softening type]        
			 const real      th,       //[I: tolerance parameter]   
			 const uint      nu,       //[I: N_reuse]               
			 const uint      nc,       //[I: N_crit]                
			 const kern_type ker,      //[I: softening kernel]      
			 const uint      nr,       //[I: N_ref]                 
			 const potential*p,        //[I: P_ex]                  
			 const uint      nd[4]     //[I: direct sum control]    
#ifdef __GRAPE
			 ,const bool     gr        //[I: use GRAPE?]            
#endif
			 ) :
  falcON(b, (ns==0 && e>=zero)? e:-abs(e), th, ker),
  BODIES(b), PEX(p), TINI(ti), C_OLD(clock()), 
  NSOFT(ns), NREF(nr), NREUSE(nu), N_REUSED(0), NCRIT(max(1u,nc)), EMAX(e), 
#ifdef __GRAPE
  GRAPE(gr),
#endif
  DIAG(false), CPU_TOTAL(zero)
{
  DIRECT[0] = nd[0];
  DIRECT[1] = nd[1];
  DIRECT[2] = nd[2];
  DIRECT[3] = nd[3];
  if(PEX) BODIES->enable_external_pot();
  reset_softening(sf,ns,e,ker);
}
//------------------------------------------------------------------------------
inline void basic_nbody::estimate_mass_density(const bool ext)
{
  register clock_t  cpu = clock();                 // cpu time                  
  falcON::estimate_rho(NREF);                      // estimate rho in tree      
  if(PEX && ext)                                   // if external potential     
    LoopMyBodies if(is_sink(Bi))                   //   loop body sinks         
      Bi.rho()+=PEX->rho(pos(Bi),time());          //     add ext density       
  record_cpu(cpu,CPU_DENS);                        // record CPU                
}
//------------------------------------------------------------------------------
void basic_nbody::estimate_mass_densities(const bool ext)
{
  LoopMyBodies Bi.flag_as_sink();
  estimate_mass_density(ext);
}
//------------------------------------------------------------------------------
void basic_nbody::estimate_surf_densities()
{
  LoopMyBodies Bi.flag_as_sink();
  falcON::estimate_sd(NREF);
}
//------------------------------------------------------------------------------
inline real basic_nbody::trace(const tensor&T) const
{
  register real tr=zero;
  for(register int i=0; i<NDIM; i++) tr += T[i][i];
  return tr;
}
//------------------------------------------------------------------------------
void basic_nbody::diagnose() const
#define LoopTensor for(i=0; i<NDIM; i++) for(j=0; j<NDIM; j++)
{
  register indx   i,j;
  register double u=zero, m=zero, t=zero, ue=zero;
  register vect   x=zero, v=zero;
  register amom   l=zero;
  register tensor T,W;
  LoopTensor T[i][j] = W[i][j] = zero;
  if(PEX) {
    register real uer=zero, mr=zero;
    PEX->energies(uer,W,T,x,v,l,mr);
    ue = uer;
    m  = mr;
    LoopMyBodies ue += mass(Bi) * pex(Bi);         // external potential energy 
  }
  LoopMyBodies {
    m += double(mass(Bi));                         // total mass                
    u += double(mass(Bi)) * pot(Bi);               // internal potential energy 
    t += double(mass(Bi)) * 
      ( square(double(vel(Bi)[0])) +
        square(double(vel(Bi)[1])) +
        square(double(vel(Bi)[2])) );
    LoopTensor {
      T[i][j] += mass(Bi) * vel(Bi)[i]*vel(Bi)[j]; // kin energy tensor         
      W[i][j] += mass(Bi) * pos(Bi)[i]*acc(Bi)[j]; // pot energy tensor         
    }
    x += mass(Bi) * pos(Bi);                       // dipole                    
    v += mass(Bi) * vel(Bi);                       // momentum                  
    l += mass(Bi) * pos(Bi) ^ vel(Bi);             // total angular momentum    
  }
  M    = m;                                        // total mass                
  Ktot = half*t;                                   // total kin energy          
  Utot = half*u+ue;                                // total pot energy          
  TU   =-half*trace(T)/trace(W);                   // virial ratio              
  L    = l;                                        // total angular momentum    
  CMX  = x/m;                                      // center of mass            
  CMV  = v/m;                                      // center of mass velocity   
  LoopTensor {
    KT[i][j] = half * T[i][j];                     // kin energy tensor         
    WT[i][j] =        W[i][j];                     // pos energy tensor         
  }
}
#undef LoopTensor
//------------------------------------------------------------------------------
void basic_nbody::update_diagnostics() const
{
  if(DIAG) return;
  diagnose();
  DIAG = true;                                     // diagnose is up-to-date   
}
//------------------------------------------------------------------------------
const real& basic_nbody::kin_energy    () const
{
  update_diagnostics();
  return Ktot;
}
//------------------------------------------------------------------------------
const real& basic_nbody::pot_energy    () const
{
  update_diagnostics();
  return Utot;
}
//------------------------------------------------------------------------------
const real  basic_nbody::total_energy  () const
{
  update_diagnostics();
  return Ktot+Utot;
}
//------------------------------------------------------------------------------
const real& basic_nbody::virial_ratio  () const
{
  update_diagnostics();
  return TU;
}
//------------------------------------------------------------------------------
const amom& basic_nbody::total_angmom  () const
{
  update_diagnostics();
  return L;
}
//------------------------------------------------------------------------------
const vect& basic_nbody::total_momentum() const
{
  update_diagnostics();
  return CMV;
}
//------------------------------------------------------------------------------
const vect& basic_nbody::center_of_mass() const
{
  update_diagnostics();
  return CMX;
}
//------------------------------------------------------------------------------
const real& basic_nbody::kin_energy    (const indx i, const indx j) const
{
  update_diagnostics();
  return KT[i][j];
}
//------------------------------------------------------------------------------
const real& basic_nbody::pot_energy      (const indx i, const indx j) const 
{
  update_diagnostics();
  return WT[i][j];
}
//------------------------------------------------------------------------------
const real  basic_nbody::as_angmom    (const indx i, const indx j) const 
{
  update_diagnostics();
#if NDIM==3
  switch(i) {
  case 0:
    switch(j) {
    case 0:  return  zero;
    case 1:  return  L[2];
    default: return -L[1];
    }  
  case 1:
    switch(j) {
    case 0:  return -L[2];
    case 1:  return  zero;
    default: return  L[0];
    }  
  default:
    switch(j) {
    case 0:  return  L[1];
    case 1:  return -L[0];
    default: return  zero;
    }  
  }
#else
  return (i==j) ? zero : real(L);
#endif
}
////////////////////////////////////////////////////////////////////////////////
// class fasstree::LeapFrogCode                                                 
////////////////////////////////////////////////////////////////////////////////
LeapFrogCode::LeapFrogCode(const bodies    *b,     // I: bodies                 
			   const uint       ns,    // I: N_soft                 
			   const real       e,     // I: eps/eps_max            
			   const real       ti,    // I: initial time           
			   const int        h0,    // I: h0                     
			   const uint       nu,    //[I: N_reuse]               
			   const vect     * pc,    //[I: center of root]        
			   const soft_type  sf,    //[I: softening type]        
			   const real       th,    //[I: tolerance parameter]   
			   const uint       nc,    //[I: N_crit]                
			   const kern_type  ker,   //[I: softening kernel]      
			   const uint       nr,    //[I: N_ref]                 
			   const potential* pp,    //[I: P_ex]                  
			   const uint       nd[4]  //[I: direct sum control]    
#ifdef __GRAPE
			   ,const bool      gr     //[I: use GRAPE?]            
#endif
			   ) :
  basic_nbody(b,ns,e,ti,sf,th,nu,nc,ker,nr,pp,nd
#ifdef __GRAPE
	      ,gr
#endif
	      ),
  LeapFrog   (h0,ti)
{
  register clock_t  cpu = clock();                 // cpu time                  
  reset_cpus();                                    // reset CPU time counters   
  if(SOFTENING==individual_adaptive)               // if individual adaptive:   
    set_softening_lengths();                       //   then set eps_i          
  LoopMyBodies Bi.flag_as_sink();                  // flag everybody as sink    
  set_tree  (pc,grow);                             // create tree               
  set_forces();                                    // compute forces            
  record_cpu(cpu,CPU_STEP);                        // record CPU time           
  update_cpu_total();                              // record accumulated time   
}
//------------------------------------------------------------------------------
void LeapFrogCode::full_step(const vect*pc)
{
  register clock_t  cpu = clock();                 // cpu time                  
  reset_cpus();                                    // reset CPU time counters   
  predict(BODIES);                                 // move = drift              
  set_tree  (pc);                                  // create tree               
  set_forces();                                    // approximate forces        
  accelerate(BODIES);                              // accelerate = kick         
  DIAG = false;                                    // diagnose out-of-date      
  record_cpu(cpu,CPU_STEP);                        // record CPU time           
  update_cpu_total();                              // record accumulated time   
}
//------------------------------------------------------------------------------
void LeapFrogCode::long_stats(ostream& to) const
{
  to<<" cpu times [sec]: "
    <<"predict build tree/ext grav  eps_i  total   accum\n                  "
    <<setw(6)<<setprecision(4)<<cpu_build()      <<" "
    <<setw(6)<<setprecision(4)<<cpu_tree_force() <<" "
    <<setw(6)<<setprecision(4)<<cpu_ext_force()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_softening()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_longstep()   <<" "
    <<setw(8)<<setprecision(6)<<cpu_total()      <<"\n";
  update_diagnostics();
  to<<" E="<<setw(8)<<setprecision(5)<<total_energy()
    <<"; -T/U="<<setw(7)<<setprecision(5)<<virial_ratio()
    <<"; L ="  
#if NDIM==3
    <<setw(7)<<setprecision(5)<<total_angmom()[0]<<" "
    <<setw(7)<<setprecision(5)<<total_angmom()[1]<<" "
    <<setw(7)<<setprecision(5)<<total_angmom()[2]
#else
    <<setw(7)<<setprecision(5)<<total_angmom()
#endif
    <<"; Vcm= "<<setw(7)<<setprecision(4)<<total_momentum()[0]<<" "
    <<setw(7)<<setprecision(4)<<total_momentum()[1]
#if NDIM==3
    <<" "<<setw(7)<<setprecision(4)<<total_momentum()[2]
#endif
    <<endl;
}
//------------------------------------------------------------------------------
void LeapFrogCode::short_stats(ostream& to) const
{
  to.setf(ios::left, ios::adjustfield);
  update_diagnostics();
  to<<setprecision(5)<<setw(10)<<time()         <<" "
    <<setprecision(7)<<setw(13)<<total_energy() <<" "
    <<setprecision(min(5,max(1,5-int(-log10(virial_ratio())))))
    <<setw(7)<<virial_ratio()<<" ";
#if NDIM==3
  to<<setprecision(5)<<setw(10)<<SQRT(norm(total_angmom()))  <<" ";
#else
  to<<setprecision(5)<<setw(10)<<total_angmom()              <<" ";
#endif
  to<<setprecision(2)<<setw(8)<<SQRT(norm(total_momentum()))  <<" ";
  to.setf(ios::right, ios::adjustfield);
  to<<setprecision(4)<<setw(6)<<cpu_build()    <<" "
    <<setprecision(4)<<setw(7)<<cpu_force()    <<" "
    <<setprecision(4)<<setw(7)<<cpu_longstep() <<" "
    <<setprecision(6)<<setw(9)<<cpu_total()    <<endl;
}
//------------------------------------------------------------------------------
void LeapFrogCode::short_stat_head(ostream& to) const
{
  to<<"    time   "
    <<"   energy     "
    <<"  -T/U  "
#if NDIM==3
    <<"   |L|     "
#else
    <<"   L_z     "
#endif
    <<"  |v_cm| "
    <<" build   force    step     accum"<<endl;
}
//------------------------------------------------------------------------------
void LeapFrogCode::short_stat_line(ostream &to) const
{
  to<<" ----------------------------------------------------"
    "--------------------------------"<<endl;
}
////////////////////////////////////////////////////////////////////////////////
// class fasstree::BlockStepCode                                                
////////////////////////////////////////////////////////////////////////////////
void BlockStepCode::reset_stepping(const real fa, const real fp, const real fc)
{
  BlockStep::reset_scheme(fa,fp,fc);
}
//------------------------------------------------------------------------------
inline void BlockStepCode::set_sink_flags(const indx low)
{
  LoopMyBodies                                     // loop bodies in tree       
    if(level(Bi) >= low) Bi.flag_as_sink();        //   IF(level>=low): sink    
    else                 Bi.unflag_sink ();        //   ELSE          : no sink 
}
//------------------------------------------------------------------------------
#define ForAll    LoopMyBodies
#define ForActive LoopMyBodies if(is_sink(Bi))
void BlockStepCode::
elementary_step(const indx low,                    // I: lowest level moving    
		const vect*x,                      //[I: center of root]        
		const how_to_make_tree make)       //[I: how to make tree]      
{
  static uint m = 0u;                              // # of tiny moves omitted   
  clock_on();                                      // increase actual clock     
  if(need_move(low)) {                             // IF anybody is sink now   >
    register real dt = tau_min() * (m+1);          //   dt = (m+1)*tau_min      
    ForAll move_by(Bi,dt);                         //   all:    x -> x+v*dt     
    m = 0;                                         //   reset m = 0             
    set_tree(x,make);                              //   initialize tree         
    set_sink_flags(low);                           //   if(l>=low) -> active    
    set_forces();                                  //   active: a = acc(x)      
    ForActive acce_half(Bi);                       //   active: v -> v+a*h/2    
    if(low != highest_level())                     //   IF(levels may change)  >
      ForActive adjust_level(Bi,low);              //     active: h -> h'      <
    if(low)                                        //   IF(not last step)      >
      ForActive acce_half(Bi);                     //     active: v -> v+a*h/2 <
  } else m++;                                      // < ELSE count move omission
}
//------------------------------------------------------------------------------
void BlockStepCode::full_step(const vect* x)
{
  register clock_t  cpu = clock();                 // cpu time                  
  const    unsigned nn  = 1 << highest_level();    // number of small steps     
  register unsigned tn;                            // index of small step       
  reset_cpus();                                    // reset CPU time counters   
  ForAll acce_half(Bi);                            // v -> v + a*h/2            
  for(tn=1; tn<nn; tn++)                           // loop over all but last    
    elementary_step(longest_moving(tn),x);         //   elementary steps        
  elementary_step(0,x,grow);                       // last step: grow tree      
  DIAG = false;                                    // diagnose out-of-date      
  record_cpu(cpu,CPU_STEP);                        // record CPU time           
  update_cpu_total();                              // record accumulated time   
}
#undef ForAll
#undef ForActive
//------------------------------------------------------------------------------
void BlockStepCode::prepare(const int  h0,         // I: h0                     
			    const indx nt,         // I: N_steps                
			    const vect*x)          //[I: center of root]        
{
  BODIES->enable_time_steps();                     // alloc levels              
  register clock_t cpu = clock();                  // local CPU counter         
  reset_cpus();                                    // reset CPU counters        
  BlockStep::reset_steps(h0,nt,TINI);              // set time steps            
  set_tree(x,grow);                                // initialize tree           
  if(SOFTENING != global_fixed)                    // if individual softening   
    set_softening_lengths();                       //   set eps_i               
  LoopMyBodies Bi.flag_as_sink();                  // flag every body as sink   
  set_forces();                                    // get accelerations         
  diagnose();                                      // energy et al. at t=t0     
  DIAG = true;                                     // diagnose up-to-date       
  BlockStep::reset_counts();                       // reset counts              
  LoopMyBodies                                     // loop bodies               
    assign_level(Bi);                              //   assign step level       
  record_cpu(cpu,CPU_STEP);                        // record total CPU          
  update_cpu_total();                              // record accumulated time   
}
//------------------------------------------------------------------------------
BlockStepCode::BlockStepCode(const bodies   *b,    // I: bodies                 
			     const uint      ns,   // I: N_soft                 
			     const real      e,    // I: eps/eps_max            
			     const real      ti,   // I: initial time           
			     const int       h0,   // I: h0                     
			     const indx      nl,   // I: N_steps                
			     const uint      nu,   // I: N_reuse                
			     const real      fa,   // I: factor f_a             
			     const real      fp,   //[I: factor f_p]            
			     const real      fc,   //[I: factor f_c]            
			     const vect     *pc,   //[I: center of root]        
			     const soft_type sf,   //[I: softening type]        
			     const real      th,   //[I: tolerance parameter]   
			     const uint      nc,   //[I: N_crit]                
			     const kern_type ker,  //[I: softening kernel]      
			     const uint      nr,   //[I: N_ref]                 
			     const potential*pp,   //[I: P_ex]                  
			     const uint      nd[4] //[I: direct sum control]    
#ifdef __GRAPE
			     ,const bool     gr    //[I: use GRAPE?]            
#endif
			     ) :
  basic_nbody(b,ns,e,ti,sf,th,nu,nc,ker,nr,pp,nd
#ifdef __GRAPE
	      ,gr
#endif
	      ),
  BlockStep  (),
  W          (max(6, int(log10(double(b->N_body())))))
{
  BlockStep::reset_scheme(fa,fp,fc);
  if(fa==zero && fp==zero) error("time step control factors = 0");
  prepare(h0,nl,pc);
}
//------------------------------------------------------------------------------
void BlockStepCode::long_stats(ostream& to) const
{
  to<<" cpu times [sec]: "
    <<"predict build tree/ext grav  eps_i  total   accum\n                  "
    <<setw(6)<<setprecision(4)<<cpu_build()      <<" "
    <<setw(6)<<setprecision(4)<<cpu_tree_force() <<" "
    <<setw(6)<<setprecision(4)<<cpu_ext_force()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_softening()  <<" "
    <<setw(6)<<setprecision(4)<<cpu_longstep()   <<" "
    <<setw(8)<<setprecision(6)<<cpu_total()      <<"\n";
  BlockStep::short_stats(to); to<<"\n";
  if(time() != TINI) update_diagnostics();
  to<<" E="<<setw(8)<<setprecision(5)<<total_energy()
    <<"; -T/U="<<setw(7)<<setprecision(5)<<virial_ratio()
    <<"; L ="  
#if NDIM==3
    <<setw(7)<<setprecision(5)<<total_angmom()[0]<<" "
    <<setw(7)<<setprecision(5)<<total_angmom()[1]<<" "
    <<setw(7)<<setprecision(5)<<total_angmom()[2]
#else
    <<setw(7)<<setprecision(5)<<total_angmom()
#endif
    <<"; Vcm= "<<setw(7)<<setprecision(4)<<total_momentum()[0]<<" "
    <<setw(7)<<setprecision(4)<<total_momentum()[1]
#if NDIM==3
    <<" "<<setw(7)<<setprecision(4)<<total_momentum()[2]
#endif
    <<endl;
}
//------------------------------------------------------------------------------
void BlockStepCode::short_stats(ostream& to) const
{
  to.setf(ios::left, ios::adjustfield);
  if(time() != TINI) update_diagnostics();
  to<<setprecision(5)<<setw(10)<<time()         <<" "
    <<setprecision(7)<<setw(13)<<total_energy() <<" "
    <<setprecision(min(5,max(1,5-int(-log10(virial_ratio())))))
    <<setw(7)<<virial_ratio()<<" ";
#if NDIM==3
  to<<setprecision(5)<<setw(10)<<SQRT(norm(total_angmom()))  <<" ";
#else
  to<<setprecision(5)<<setw(10)<<total_angmom()              <<" ";
#endif
  to<<setprecision(2)<<setw(8)<<SQRT(norm(total_momentum()))  <<" ";
  to.setf(ios::right, ios::adjustfield);
  if(highest_level()) BlockStep::short_stats(to,W);
  to<<setprecision(4)<<setw(6)<<cpu_build()    <<" "
    <<setprecision(4)<<setw(7)<<cpu_force()    <<" "
    <<setprecision(4)<<setw(7)<<cpu_longstep() <<" "
    <<setprecision(6)<<setw(9)<<cpu_total()    <<endl;
}
//------------------------------------------------------------------------------
namespace nbdy {
  inline ostream& put_char(ostream&o, const char c, const int s)
  {
    if(s>0) for(register int i=0; i<s; ++i) o<<c;
    return o;
  }
}
//------------------------------------------------------------------------------
void BlockStepCode::short_stat_head(ostream& to) const
{
  to<<"    time   "
    <<"   energy     "
    <<"  -T/U  "
#if NDIM==3
    <<"   |L|     "
#else
    <<"    L      "
#endif
    <<"  |v_cm| ";
  if(highest_level())
    for(register int i=0, h=-h0(); i<=highest_level(); i++, h--)
           if(h>13)  put_char(to,' ',W-5)<<" 2^" <<     h  <<" ";
      else if(h> 9)  put_char(to,' ',W-5)<<" "   << (1<<h) <<" ";
      else if(h> 6)  put_char(to,' ',W-5)<<"  "  << (1<<h) <<" ";
      else if(h> 3)  put_char(to,' ',W-5)<<"   " << (1<<h) <<" ";
      else if(h>=0)  put_char(to,' ',W-5)<<"    "<< (1<<h) <<" ";
      else if(h==-1) put_char(to,' ',W-5)<<"  1/2 ";
      else if(h==-2) put_char(to,' ',W-5)<<"  1/4 ";
      else if(h==-3) put_char(to,' ',W-5)<<"  1/8 ";
      else if(h==-4) put_char(to,' ',W-5)<<" 1/16 ";
      else if(h==-5) put_char(to,' ',W-5)<<" 1/32 ";
      else if(h==-6) put_char(to,' ',W-5)<<" 1/64 ";
      else if(h>-10) put_char(to,' ',W-5)<<" 2^" <<     h  <<" ";
      else           put_char(to,' ',W-5)<<"2^"  <<     h  <<" ";
  to<<" build   force    step     accum"<<endl;
}
//------------------------------------------------------------------------------
void BlockStepCode::short_stat_line(ostream &to) const
{
  to<<" ----------------------------------------------------";
  if(highest_level())
    for(register int i=0; i<=highest_level(); i++) put_char(to,'-',W+1);
  to<<"--------------------------------"<<endl;
}
//------------------------------------------------------------------------------
#undef LoopMyBodies
